import sys
import importlib

def parse(params, create_module):
    for lib in params['import']:
        module = importlib.import_module(lib)
        globals()[lib] = module

    exec(params['code'], globals())
    if result is None:
        raise Exception("The code argument did not assign a value to 'result' variable")
    return result



import torch
from scalenet.residuals import combine_residuals

class SiameseNet(torch.nn.Module):
    def __init__(self, net):
        super().__init__()
        self.module = net

    def forward(self, x_list):
        return [self.module(x) for x in x_list]

def parse(params, create_module):
    net = params['arch_desc']['net']
    siam = SiameseNet(create_module(net))
    return siam



import torch
from scalenet import UpsampleResiduals
from scalenet.residuals import res_warp_img, combine_residuals

class HindsightWarp(torch.nn.Module):
    def __init__(self, rollback_range, inner_module):
        super().__init__()
        self.rollback_range = rollback_range
        self.inner_module = inner_module

    def forward(self, x, res, state, level):
        #Maybe this exit condition shouldn't be appplied to hindsight case
        if len(torch.nonzero(res))  == 0: #res is all 0's
            return x

        res = res.permute(0, 2, 3, 1)

        num_fms = x.shape[1]
        src = x[:, :num_fms//2]
        tgt = x[:, num_fms//2:]
        warped_tgt = res_warp_img(tgt, res, is_pix_res=True, rollback=self.rollback_range)
        res = res.permute(0, 3, 1, 2)

        in_bundle = torch.cat((src, tgt, warped_tgt, res), 1)

        hindsight_adjustment = self.inner_module(in_bundle)

        # this is aweful
        state['down'][str(level)]['input'] =  combine_residuals(res.permute(0, 2, 3, 1),
                                    hindsight_adjustment.permute(0, 2, 3, 1),
                                    rollback=self.rollback_range).permute(0, 3, 1, 2)

        res = state['down'][str(level)]['input'].permute(0, 2, 3, 1)
        hindsight_warped_tgt = res_warp_img(tgt, res, is_pix_res=True, rollback=self.rollback_range)
        result = torch.cat((src, hindsight_warped_tgt), 1)

        return result

def parse(params, create_module):
    rollback_range = params['arch_desc']['rollback']
    inner_module = create_module(params['arch_desc']['inner_module'])
    net = HindsightWarp(rollback_range, inner_module)
    return net



import torch

def parse(params, create_module):
    if 'scale' in params:
        scale = params['scale']
    else:
        scale = 2

    if 'mode' in params:
        mode = params['mode']
    else:
        mode = 'bilinear'

    return torch.nn.Upsample(scale_factor=scale, mode=mode)



import torch
from scalenet import Model

class ModelWithWeights(Model):
    def __init__(self, m, num_weights):
        super().__init__()
        self.m = m
        self.weights = torch.nn.Parameter(torch.cuda.FloatTensor([1.0 for i in range(num_weights)]))

    def forward(self, x):
        return self.m(x)

def parse(params, create_module):
    layer_list = []
    model = create_module(params['model'])
    num_weights = params['num_weights']
    net = ModelWithWeights(model, num_weights)
    return net



import torch
from scalenet import UpsampleResiduals
from scalenet.residuals import res_warp_img

class RollbackWarp(torch.nn.Module):
    def __init__(self, rollback_range):
        super().__init__()
        self.rollback_range = rollback_range
        self.ups_res = UpsampleResiduals()
        self.ups_img = torch.nn.Upsample(scale_factor=2, mode='bilinear')
        self.downs_img = torch.nn.AvgPool2d(2)

    def forward(self, x, res, state, level):
        if len(torch.nonzero(res))  == 0: #res is all 0's
            return x

        res = res.permute(0, 2, 3, 1)

        #TODO: test whether this is good enough
        #for _ in range(self.rollback_range):
        #    x = self.ups_img(x)
        #    res  = self.ups_res(res)
        #import pdb; pdb.set_trace()
        num_fms = x.shape[1]
        src = x[:, :num_fms//2]
        tgt = x[:, num_fms//2:]
        warped_tgt = res_warp_img(tgt, res, is_pix_res=True, rollback=self.rollback_range)
        result = torch.cat((src, warped_tgt), 1)
        #for _ in range(self.rollback_range):
        #    result = self.downs_img(result)

        return result

def parse(params, create_module):
    rollback_range = params['arch_desc']['rollback']
    net = RollbackWarp(rollback_range)
    return net



import torch
import copy

class MaskedNormalize(torch.nn.Module):
    def __init__(self, mask_value=0, mask_type='eq', mask_fill=None,
                per_channel_mean=True, per_channel_var=False):
        super().__init__()
        self.mask_value = mask_value
        self.mask_type = mask_type
        self.mask_fill = mask_fill
        self.per_channel_mean = per_channel_mean
        self.per_channel_var = per_channel_var

    def get_mask(self, x):
        if self.mask_type == 'eq':
            return x == self.mask_value
        elif self.mask_type == 'lt':
            return x < self.mask_value
        elif self.mask_type == 'gt':
            return x > self.mask_value
        elif self.mask_type == 'lte':
            return x <= self.mask_value
        elif self.mask_type == 'gte':
            return x >= self.mask_value
        else:
            raise Exception("Maks type unknown: {}".format(mask_type))

    def forward(self, x, eps=1e-5):
        x_out = x.clone()
        mask = self.get_mask(x) == False
        for b in range(x.shape[0]):
            x = x_out[b]
            if self.per_channel_mean and len(x.shape) == 4:
                for f in range(x.shape[1]):
                    m = mask[b, f]
                    x[f][m] = x[f][m].clone() - torch.mean(x[f][m].clone())
            else:
                m = mask[b]
                x[m] = x[m].clone() - torch.mean(x[m].clone())

            if self.per_channel_var and len(x.shape) == 4:
                for f in range(x.shape[1]):
                    m = mask[b, f]
                    var = torch.var(x[f][m].clone())
                    x[f][m] = x[f][m].clone() / (torch.sqrt(var) + eps)
            else:
                m = mask[b]
                var = torch.var(x[m].clone())
                x[m] = x[m].clone() / (torch.sqrt(var) + eps)

        if self.mask_fill is not None:
            x_out[mask == False] = self.mask_fill

        return x_out

def parse(params, create_module):
    params_tmp = copy.deepcopy(params)
    del params_tmp['type']
    net = MaskedNormalize(**params_tmp)
    return net



import torch
from scalenet.residuals import combine_residuals

class FmRangeApply(torch.nn.Module):
    def __init__(self, module, fm_range):
        super().__init__()
        self.module = module
        self.fm_range = fm_range

    def forward(self, x):
        b_in, c_in, *rest_in = x.shape
        assert c_in % self.fm_range == 0
        x_per_fm = x.view(b_in * (c_in // self.fm_range), self.fm_range, *rest_in)
        out_per_fm = self.module(x_per_fm)
        b_out, c_out, *rest_out = out_per_fm.shape
        out = out_per_fm.view(b_in, -1, *rest_out)

        return out

def parse(params, create_module):
    inner_module = params['module']
    fm_range = params['fm_range']
    net = FmRangeApply(create_module(inner_module), fm_range)
    return net



import torch

def parse(params, create_module):
    layer_list = []
    modules = [create_module(m) for m in params['arch_desc']['module_list']]

    net = torch.nn.Sequential(*modules)
    return net



import scalenet
import six

def parse(params, create_module):
    if 'max_level' in params:
        net = scalenet.ScaleNet(max_level=params['max_level'])
    else:
        net = scalenet.ScaleNet()

    if "downmodules" in params:
        for level, spec in six.iteritems(params["downmodules"]):
            module = create_module(spec)
            net.set_downmodule(module, level)

    if "upmodules" in params:
        for level, spec in six.iteritems(params["upmodules"]):
            module = create_module(spec)
            net.set_upmodule(module, level)

    if "uplinks" in params:
        if "all" in params["uplinks"]:
            spec = params["uplinks"]["all"]
            module = create_module(spec)
            net.set_all_uplinks(module)
        else:
            for level, spec in six.iteritems(params["uplinks"]):
                module = create_module(spec)
                net.set_uplink(module, level)

    if "downlinks" in params:
        if "all" in params["downlinks"]:
            spec = params["downlinks"]["all"]
            module = create_module(spec)
            net.set_all_downlinks(module)
        else:
            for level, spec in six.iteritems(params["downlinks"]):
                module = create_module(spec)
                net.set_downlink(module, level)

    if "skiplinks" in params:
        if "all" in params["skiplinks"]:
            spec = params["skiplinks"]["all"]
            module = create_module(spec)
            net.set_all_skiplinks(module)
        else:
            for level, spec in six.iteritems(params["skiplinks"]):
                module = create_module(spec)
                net.set_skiplink(module, level)

    if "combiners" in params:
        if "all" in params["combiners"]:
            spec = params["combiners"]["all"]
            module = create_module(spec)
            net.set_all_combiners(module)
        else:
            for level, spec in six.iteritems(params["combiners"]):
                module = create_module(spec)
                net.set_combiner(module, level)

    return net



import json
import glob
import os
from importlib import util
import torch

def import_file(full_name, path):
    spec = util.spec_from_file_location(full_name, path)
    mod = util.module_from_spec(spec)

    spec.loader.exec_module(mod)
    return mod

def find(name, path):
    for root, dirs, files in os.walk(path):
        if name in files:
            return os.path.join(root, name)

def find_file_in_folder_set(filename, folder_set):
    filename = os.path.expanduser(filename)
    found_file = None

    if os.path.isfile(filename):
        found_file = filename
    else:
        global_path = None
        for parent in folder_set:
            global_path = os.path.join(parent, filename)

            if os.path.isfile(global_path):
                found_file = global_path
                break
    return found_file

class Artificery():
    def __init__(self, checkpoint_init=False):
        self.reload_parsers()
        self.param_folders = set()
        self.used_specfiles = []
        self.checkpoint_init = checkpoint_init

    def parse(self, params_file):
        params_file = os.path.expanduser(params_file)
        print (params_file)

        folder = os.path.dirname(params_file)
        if folder != '' and folder not in self.param_folders:
            added_folder = True
            self.param_folders.add(folder)
        else:
            added_folder = False

        found_file = find_file_in_folder_set(params_file, self.param_folders)
        if found_file is None:
            raise Exception("Cannot find params file: {}".format(params_file))

        with open(found_file, 'r') as f:
            params = json.load(f)
        self.used_specfiles.append(found_file)

        net = self.create_net(params)
        if added_folder:
            self.param_folders.remove(folder)

        return net

    def create_net(self, params):
        if 'path' in params:
            return self.parse(params['path'])

        elif 'type' in params:
            net_type = params['type'].lower()
            net = self.parsers[net_type](params, create_module=self.create_net)
        else:
            raise Exception("Neither type nor path specified.")

        if self.checkpoint_init and 'checkpoint_init' in params:
            checkpoint_file = params['checkpoint_init']
            found_file = find_file_in_folder_set(checkpoint_file, self.param_folders)
            if found_file is None:
                raise Exception("Cannot find init file: {}".format(checkpoint_file))
            print ("loading weights from {}".format(found_file))
            net.load_state_dict(torch.load(found_file))
            print ("loaded! mean weight [0] == {}".format(torch.mean(list(net.parameters())[0])))
        if 'trainable' in params and params['trainable'] == False:
            print ("Setting layer to non trainable")
            for param in net.parameters():
                param.requires_grad = False
        return net

    def reload_parsers(self):
        parser_template = os.path.join(os.path.dirname(__file__), "parsers/*")
        parser_folders = glob.glob(parser_template)

        self.parsers = {}
        for folder in parser_folders:
            if os.path.isdir(folder):
                parser_path = os.path.join(folder, "parse.py")
                if os.path.isfile(parser_path) :
                    parser_name = folder.split('/')[-1]

                    if parser_name.lower() != parser_name:
                        raise Exception("Parser name cannot contain uppercase letters."
                                        "Violator: {}".format(folder))

                    parser  = import_file("parser", parser_path)
                    self.parsers[parser_name] = parser.parse





import torch
from scalenet import Model

class TransparentSeq(Model):
    def __init__(self, stages):
        super().__init__()
        self.stages = torch.nn.ModuleList(stages)
        self.inermediate = []

    def forward(self, x, start_stage=None, end_stage=None):
        self.intermediate = []
        if start_stage is None:
            start_stage = 0
        if end_stage is None:
            end_stage = len(self.stages)

        for i in range(start_stage, end_stage):
            m = self.stages[i]
            x = m(x)
            self.intermediate.append(x)

        return self.intermediate[-1]

def parse(params, create_module):
    layer_list = []
    module_specs = params["modules"]

    modules = [create_module(m) for m in params['modules']]

    net = TransparentSeq(modules)
    return net



import torch
import copy

class MaskOutputBasedOnInput(torch.nn.Module):
    def __init__(self, inner_module, mask_value=0, mask_type='eq', mask_fill=0):
        super().__init__()
        self.mask_value = mask_value
        self.mask_type = mask_type
        self.mask_fill = mask_fill
        self.inner_module = inner_module

    def get_mask(self, x):
        if self.mask_type == 'eq':
            return x == self.mask_value
        elif self.mask_type == 'lt':
            return x < self.mask_value
        elif self.mask_type == 'gt':
            return x > self.mask_value
        elif self.mask_type == 'lte':
            return x <= self.mask_value
        elif self.mask_type == 'gte':
            return x >= self.mask_value
        else:
            raise Exception("Maks type unknown: {}".format(mask_type))

    def downsample(self, x, target_shape):
        assert len(target_shape) == 4
        assert len(x.shape) == 4
        assert target_shape[0:2] == x.shape[0:2]
        assert target_shape[2] == target_shape[3]
        assert x.shape[2] == x.shape[3]
        assert x.shape[2] % target_shape[2] == 0
        ratio = x.shape[2] // target_shape[2]

        result = x.clone()
        result = torch.nn.functional.max_pool2d(result, ratio)

        if result.shape != target_shape:
            import pdb; pdb.set_trace()
            raise Exception("Can't downsample {} to {}".format(x.shape, target_shape))
        return result

    def forward(self, x):
        y = self.inner_module(x)
        x_mask = self.get_mask(x)
        y_mask = self.downsample(x_mask.float(), y.shape).byte()
        y[y_mask] = self.mask_fill
        return y

def parse(params, create_module):
    inner_module = create_module(params['inner_module'])
    params_tmp = copy.deepcopy(params)
    del params_tmp['type']
    del params_tmp['inner_module']
    net = MaskOutputBasedOnInput(inner_module=inner_module, **params_tmp)
    return net



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="artificery",
    version="0.0.2.8",
    author="Sergiy Popovych",
    author_email="sergiy.popovich@gmail.com",
    description="A tool for building pyrorch nets from spec files",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/supersergiy/artificery",
    include_package_data=True,
    package_data={'': ['*.py']},
    install_requires=[
      'scalenet',
      'torch',
      'torchvision'
    ],
    packages=setuptools.find_packages(),
)



import torch

class Vectorizer(torch.nn.Module):
    def __init__(self, out_channels=2, component_channels=7,
                 grid=False, norm_method='softmax',
                 permute=True, step=None,
                 max_value=None,
                 train_step=False):
        super(Vectorizer, self).__init__()

        if step is not None and max_value is not None:
            raise Exception("Ambiguous specification for categorical regression: "
                    "both step and max_value are specified")
        if step is None:
            if max_value is None:
                step = 1
            else:
                step = max_value / (component_channels//2)

        self.out_channels = out_channels
        self.component_channels = component_channels
        self.norm_method = norm_method
        self.permute = permute
        self.grid = grid

        if self.grid:
            x, y = torch.meshgrid([torch.range(0, self.component_channels - 1, device='cuda',
                                 requires_grad=False)]*2
                                 )
            x = x - self.component_channels//2
            y = y - self.component_channels//2

            self.weights = torch.cat((x.unsqueeze(-1), y.unsqueeze(-1)), 2)
            self.weights = self.weights.view(-1, 2)
        else:
            self.weights = torch.range(0, self.component_channels - 1, device='cuda',
                                  requires_grad=False) - self.component_channels//2
            self.weights = self.weights.unsqueeze(-1)

        self.weights *= step

        if train_step:
            self.weights = torch.nn.Parameter(self.weights)

        norm_methods = {
            'softmax': torch.nn.Softmax(-1),
            'softmin': torch.nn.Softmin(-1),
            'norm': torch.nn.LayerNorm(self.component_channels,
                                       eps=1e-05, elementwise_affine=False)
        }
        self.normer = norm_methods[self.norm_method]

    def forward(self, x_in):
        if self.permute:
            x = x_in.permute(0, 2, 3, 1)
        else:
            x = x_in

        if self.grid:
            result = self.grid_forward(x)
        else:
            result = self.channelwise_forward(x)

        if self.permute:
            result = result.permute(0, 3, 1, 2)
        return result

    def channelwise_forward(self, x):
        components = []
        norm_components = []
        result_components = []
        for c_id in range(self.out_channels):
            ch_start = c_id * self.component_channels
            ch_end = (c_id + 1) * self.component_channels
            component = x[..., ch_start:ch_end]
            #print ('Component: ', torch.mean(component))
            norm_component = self.normer(component)
            norm_components.append(norm_component)
            norm_component_2d = norm_component.view(-1, self.component_channels)
            #print (norm_component)
            component_result_2d = torch.mm(norm_component_2d, self.weights)
            component_result = component_result_2d.view(list(component.shape[0:-1]) + [1])
            result_components.append(component_result)

        result = torch.cat(result_components, -1)
        #print ('Result: ', result.shape)
        #print ('Result: ', torch.mean(result))
        return result

    def grid_forward(self, x):
        components = []
        norm_components = []
        result_components = []

        ch_start = 0
        ch_end = self.component_channels**2
        component = x[..., ch_start:ch_end]

        #print ('Component: ', torch.mean(component))
        norm_component = self.normer(component)
        norm_components.append(norm_component)
        norm_component_2d = norm_component.view(-1, self.component_channels**2)
        #print (norm_component)
        component_result_2d = torch.mm(norm_component_2d, self.weights)
        component_result = component_result_2d.view(list(component.shape[0:-1]) + [self.out_channels])
        result_components.append(component_result)

        result = torch.cat(result_components, -1)
        #print ('Result: ', result.shape)
        #print ('Result: ', torch.mean(result))
        return result

def parse(params, create_module):
    params = params['arch_desc']
    net = Vectorizer(**params)
    return net



import torch
from scalenet.residuals import combine_residuals, upsample_residuals

class CombinInOutRes(torch.nn.Module):
    def __init__(self, rollback_range):
        super().__init__()

    def forward(self, x, res, level, up_result):
        #TODO: test whether this is good enough
        for _ in range(self.rollback_range):
            x = self.ups_img(x)
            res  = self.ups_res(res)

        result = res_warp_img(x, res, is_pix_res=True)

        for _ in range(self.rollback_range):
            result = self.downs_img(result)

        return result

def parse(params, create_module):
    rollback_range = params['arch_desc']['rollback']
    net = RollbackWarp(rollback_range)
    return net



import torch
from scalenet.residuals import combine_residuals

class WarpWithInputDownmodule(torch.nn.Module):
    def __init__(self, inner_module):
        super().__init__()
        self.module = inner_module

    def forward(self, x, state, level):
        my_in = state['down'][str(level)]['input']
        my_out = self.module(x)
        result = combine_residuals(my_out.permute(0, 2, 3, 1),
                                   my_in.permute(0, 2, 3, 1), is_pix_res=True).permute(0, 3, 1, 2)

        return result

def parse(params, create_module):
    inner_module = params['arch_desc']['inner_module']
    net = WarpWithInputDownmodule(create_module(inner_module))
    return net
